#include <stdio.h>
#include <stdlib.h>
#include "mother.h"

const char* __asan_default_options()
{
	return "detect_leaks=0";
}

void WriteWords(FILE* fp, unsigned int v, unsigned int count)
{
	for(unsigned int i = 0; i < count; i++, v >>= 8)
		fputc(v & 0xff, fp);
}

void WriteWave(const char* filename, signed short* data, unsigned int size)
{
	FILE* fp = fopen(filename, "wb");
	
	fputs("RIFF", fp);
	WriteWords(fp, 36 + (size * 2), 4);
	fputs("WAVEfmt ", fp);
	WriteWords(fp, 16, 4);
	WriteWords(fp, 1, 2);
	WriteWords(fp, 2, 2);
	WriteWords(fp, 44100, 4);
	WriteWords(fp, 176400, 4);
	WriteWords(fp, 4, 2);
	WriteWords(fp, 16, 2);
	fputs("data", fp);
	WriteWords(fp, size * 2, 4);
	fwrite(data, size * sizeof(signed short), 1, fp);
	fclose(fp);
}

void Progress(float p)
{
	printf("Rendering: %.2f%%    \r", p * 100.0);
	fflush(NULL);
}

int main(int argc, const char** argv)
{
    if(argc < 3)
    {
        printf("Usage: mother <input bytestream> <output.wav>\n\n");
        return 1;
    }

    FILE* fp = fopen(argv[1], "rb");

    if(fp == NULL)
	{
		printf("ERROR: Could not read to specified input file '%s'.\n", argv[1]);
		return 1;
	}

	fseek(fp, 0, SEEK_END);
    long size = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    unsigned char* bytestream = malloc(size);

    fread(bytestream, size, 1, fp);
    fclose(fp);

    AudioBuffer* audio = Mother(bytestream, Progress);

	printf("                    \r");
    WriteWave(argv[2], audio->data, audio->size);

    return 0;
}