#ifndef INCLUDED_MOTHER
#define INCLUDED_MOTHER

#define ST_CONST 0
#define ST_ENV 1
#define ST_OSC 2
#define ST_MIX 3

#define TC_NOTE_ON 0
#define TC_NOTE_OFF 1

#define FT_LP6 0
#define FT_LP12 1
#define FT_LP18 2
#define FT_LP24 3
#define FT_HP6 4
#define FT_HP12 5
#define FT_HP18 6
#define FT_HP24 7
#define FT_BP12 8
#define FT_BP24 9
#define FT_NOTCH12 10
#define FT_NOTCH18 11
#define FT_AP12 12
#define FT_AP24 13
#define FT_PHASER12 14
#define FT_PHASER24 15

#define ENV_NONE 0
#define ENV_LINEAR 1
#define ENV_COSINE 2

#define OMOP_ADD 0
#define OMOP_MUL 1
#define OMOP_DIV 2
#define OMOP_COS 3
#define OMOP_MIN 4
#define OMOP_MAX 5
#define OMOP_MAP 6
#define OMOP_COMPRESS 7
#define OMOP_INST_TIME 8
#define OMOP_PAT_TIME 9
#define OMOP_PITCH 10

#define OT_FILTER 0
#define OT_SAMPLER 1
#define OT_NOISE 2
#define OT_SINE 3
#define OT_SAWTOOTH 4
#define OT_TRIANGLE 5
#define OT_PARABOLIC 6
#define OT_SQUARE 7

#define CMD_NOTE_OFF 0
#define CMD_NOTE_ON 1
#define CMD_SET_OSC_TYPE 2
#define CMD_SET_FLT_TYPE 3
#define CMD_SET_OSC_PROP 4
#define CMD_SET_FLT_PROP 5
#define CMD_SET_TRK_PROP 6
#define CMD_SET_DFLT_TYPE 7
#define CMD_HALT 8
#define CMD_OFF 9
#define CMD_SET_PARAM 10
#define CMD_SET_TEMPO 11

#define OSC_PROP_FREQ 0
#define OSC_PROP_AMP 1
#define OSC_PROP_OFFSET 2
#define OSC_PROP_PHASE 3
#define OSC_PROP_PWM 4
#define OSC_PROP_FREQMOD 5
#define OSC_PROP_WF 6

#define FLT_PROP_GAIN 0
#define FLT_PROP_FREQ 1
#define FLT_PROP_RES 2
#define FLT_PROP_DWMIX 3

#define TRK_PROP_VOL 0
#define TRK_PROP_PAN 1
#define TRK_PROP_DETUNE 2
#define TRK_PROP_DRY 3
#define TRK_PROP_DELAY 4
#define TRK_PROP_FEEDBACK 5
#define TRK_PROP_D_GAIN 6
#define TRK_PROP_D_FREQ 7
#define TRK_PROP_D_RES 8
#define TRK_PROP_D_MIX 9

#define EC_1 0
#define EC_0 1
#define EC_1000 2
#define EC_05 3
#define EC_120 4
#define EC_COUNT 5

#define CLIP_SCALE(v) (signed short)lrintf((v < -1.0f ? -1.0f : v > 1.0f ? 1.0f : v) * 32767.0f)

typedef struct 
{
    unsigned int type;
    unsigned int is_ref;
    unsigned int index;
} Sampler;

typedef struct
{
    unsigned int size;
    unsigned int intp;
    unsigned int idx;
    int sust_start;
    int sust_end;
    float rel_time;
    float act_time;
    float* tcs;
    Sampler* values;
} Envelope;

typedef struct
{
    float lv, lo, g;
} RCFilter;

typedef struct 
{
    unsigned int type;
    Sampler gain;
    Sampler freq;
    Sampler res;
    Sampler dw_mix;
    RCFilter rcf[4];
    float freq_v, res_v, g, gr, d;
} Filter;

typedef struct 
{
    unsigned int type;
    unsigned int lock;
    Sampler freq;
    Sampler amp;
    Sampler offset;
    Sampler phase;
    Sampler pwm;
    Sampler freqmul;
    Sampler waveform;
    float bf, p, lv;
    unsigned int f_count, f_mode;
    Filter* filters;
} Osc;

typedef struct 
{
    unsigned int op;
    unsigned int count;
    Sampler* elements;
} Mixer;

typedef struct
{
    unsigned int row;
    unsigned int type;
    signed int inst;
    signed int pitch;
    unsigned int prop_type;
    Sampler sampler;
    Osc* osc;
    Filter* filter;
} Command;

typedef struct 
{
    unsigned int tidx;
    Sampler value;
} RecvElem;

typedef struct 
{
    unsigned int track_idx;
    unsigned int inherits;
    unsigned int replace;
    unsigned int r_count;
    RecvElem* r_elems;
    unsigned int cmd_count;
    Command* commands;
} CmdBuf;

typedef struct
{
    Command* commands[256];
    int cidx;
} CmdList;

typedef struct 
{
    unsigned int inherits;
    unsigned int row_count;
    CmdBuf** tracks;
    unsigned int* row_frames;
} Pattern;

typedef struct 
{
    Sampler vol;
    Sampler pan;
    Sampler detune;
    Sampler dry;
    Sampler delay[16];
    Sampler feedback[16];
    Sampler params[44];
    Osc* oscs;
    float* d_buf;
    unsigned int d_idx, d_size;
    Filter* filter[16];
    int active_inst, off;
    float st, pitch, ly[16];
} Track;

typedef struct
{
    unsigned int track_count;
    unsigned int dp_count;
    unsigned int inst_count;
    unsigned int const_count;
    unsigned int env_count;
    unsigned int osc_count;
    unsigned int mix_count;
    unsigned int order_count;
    unsigned int pattern_count;
    float* constants;
    Envelope* envelopes;
    Osc* oscs;
    Mixer* mixers;
    Sampler* instruments;
    int * order;
    Pattern* patterns;
    Track* tracks;
    Sampler tempo;
    float* t_samps;
    float pt;
    float detune;
    float abst;
    unsigned int tidx;
} Score;

typedef struct
{
    int tidx;
    float duration;
} DelayParam;

// 8 seconds, stereo
#define AC_SIZE 705600

typedef struct AudioChunk AudioChunk;

struct AudioChunk
{
    unsigned int idx;
    unsigned short* data;
    AudioChunk* next;
};

typedef struct
{
    unsigned int size;
    signed short* data;
    Score* score;
} AudioBuffer;

void ProcessRow(Score* s, CmdList* cl, Pattern* cp, unsigned int ti, unsigned int cr);
AudioBuffer* Mother(const unsigned char* bytestream, void (*pg_cb)(float));

#endif // INCLUDED_MOTHER